unit Unit1;

{$mode objfpc}{$H+}

interface

uses
  Classes, SysUtils, FileUtil, Forms, Controls, Graphics, Dialogs, StdCtrls,
  Spin;

type

  { TForm1 }

  TForm1 = class(TForm)
    Button1: TButton;
    Button2: TButton;
    CheckBox1: TCheckBox;
    SpinEdit1: TSpinEdit;
    procedure Button1Click(Sender: TObject);
    procedure Button2Click(Sender: TObject);
    procedure FormClose(Sender: TObject; var CloseAction: TCloseAction);
  private
    { private declarations }
    function CreateDemoForm(StdCtrls: Boolean): TForm;
  public
    { public declarations }
  end;

  TDSEdit = class(TEdit)
  private
    FDefaultWidthInChars: Integer;
  protected
    procedure CalculatePreferredSize(var PreferredWidth,
       PreferredHeight: integer; WithThemeSpace: Boolean); override;
  published
    property DefaultWidthInChars: Integer read FDefaultWidthInChars write FDefaultWidthInChars;
  end;

  TDSMemo = class(TMemo)
  private
    FDefaultWidthInChars: Integer;
    FDefaultHeightInLines: Integer;
  protected
    procedure CalculatePreferredSize(var PreferredWidth,
       PreferredHeight: integer; WithThemeSpace: Boolean); override;
  published
    property DefaultWidthInChars: Integer read FDefaultWidthInChars write FDefaultWidthInChars;
    property DefaultHeightInLines: Integer read FDefaultHeightInLines write FDefaultHeightInLines;
  end;

var
  Form1: TForm1;

implementation

uses
  LCLIntf, LCLType;

{$R *.lfm}

function DSGetTextWidth(AChars: Integer; AFont: TFont): Integer;
var
  Bmp: TBitmap;
begin
  Result := 8;
  Bmp := TBitmap.Create;
  try
    Bmp.Canvas.Font.Assign(AFont);
    Result := (Bmp.Canvas.TextWidth('XM') div 2) * (AChars + 1);
  finally
    Bmp.Free;
  end;
end;

function DSGetTextHeight(AFont: TFont): Integer;
var
  Bmp: TBitmap;
begin
  Result := 8;
  Bmp := TBitmap.Create;
  try
    Bmp.Canvas.Font.Assign(AFont);
    Result := Bmp.Canvas.TextHeight('XM');
  finally
    Bmp.Free;
  end;
end;

procedure TDSEdit.CalculatePreferredSize(var PreferredWidth,
  PreferredHeight: integer; WithThemeSpace: Boolean);
begin
  inherited CalculatePreferredSize(PreferredWidth, PreferredHeight,
    WithThemeSpace);
  if FDefaultWidthInChars > 0 then
    PreferredWidth := DSGetTextWidth(FDefaultWidthInChars, Font);
end;

procedure TDSMemo.CalculatePreferredSize(var PreferredWidth,
  PreferredHeight: integer; WithThemeSpace: Boolean);
begin
  inherited CalculatePreferredSize(PreferredWidth, PreferredHeight,
    WithThemeSpace);
  if FDefaultWidthInChars > 0 then
    PreferredWidth := DSGetTextWidth(FDefaultWidthInChars, Font); {+ GetSystemMetrics(SM_CYHSCROLL) // + Scrollbar?}
  if FDefaultHeightInLines > 0 then
    PreferredHeight := (FDefaultHeightInLines + 1) * DSGetTextHeight(Font);
end;

{ TForm1 }

procedure TForm1.Button1Click(Sender: TObject);
begin
  with CreateDemoForm(True) do
    Show;
end;

procedure TForm1.Button2Click(Sender: TObject);
begin
  with CreateDemoForm(False) do
    Show;
end;

procedure TForm1.FormClose(Sender: TObject; var CloseAction: TCloseAction);
begin
  CloseAction := caFree;
end;

function TForm1.CreateDemoForm(StdCtrls: Boolean): TForm;
var
  E1, E2: TEdit;
  M: TMemo;
begin
  Result := TForm.Create(nil);
  if CheckBox1.Checked then
    Result.Font.Size := SpinEdit1.Value;
  Result.AutoSize := True;
  if StdCtrls then
  begin
    Result.Caption := 'Standard contols';
    E1 := TEdit.Create(Result);
    E2 := TEdit.Create(Result);
    M := TMemo.Create(Result);
  end
  else
  begin
    Result.Caption := 'Default size controls';
    E1 := TDSEdit.Create(Result);
    TDSEdit(E1).DefaultWidthInChars := 10;
    E2 := TDSEdit.Create(Result);
    TDSEdit(E2).DefaultWidthInChars := 4;
    M := TDSMemo.Create(Result);
    TDSMemo(M).DefaultWidthInChars := 10;
    TDSMemo(M).DefaultHeightInLines := 4;
  end;
  E1.Parent := Result;
  E1.AnchorSideLeft.Control := Result;
  E1.AnchorSideTop.Control := Result;
  E2.Parent := Result;
  E2.AnchorSideLeft.Control := E1;
  E2.AnchorSideLeft.Side := asrRight;
  E2.AnchorSideTop.Control := Result;
  M.Parent := Result;
  M.AnchorSideLeft.Control := Result;
  M.AnchorSideTop.Control := E1;
  M.AnchorSideTop.Side := asrBottom;
  E1.AutoSize := True;
  E2.AutoSize := True;
  M.AutoSize := True;
  Result.OnClose := @FormClose;
  E1.Text := 'ABCDEFGHIJ';
  E2.Text := 'ABCD';
  M.Append('1ABCDEFGHIJ');
  M.Append('2ABCDEFGHIJ');
  M.Append('3ABCDEFGHIJ');
  M.Append('4ABCDEFGHIJ');
end;

end.

